@@ -3,14 +3,16 @@ Data classes and utilities used by :class:`iminuit.Minuit`.
 
 You can look up the interface of data classes that iminuit uses here.
 """
+
 from __future__ import annotations
 import inspect
 from collections import OrderedDict
 from argparse import Namespace
 from iminuit import _repr_html, _repr_text, _deprecated
-from iminuit.typing import Key, UserBound
+from iminuit.typing import Key, UserBound, Cost, CostGradient
+from iminuit.warnings import IMinuitWarning, HesseFailedWarning, PerformanceWarning
 import numpy as np
-from numpy.typing import NDArray
+from numpy.typing import NDArray, ArrayLike
 from typing import (
     overload,
     Any,
@@ -23,16 +25,18 @@ from typing import (
     Optional,
     Callable,
     Collection,
+    Sequence,
+    TypeVar,
+    Annotated,
+    get_args,
+    get_origin,
 )
 import abc
 from time import monotonic
 import warnings
 import sys
 
-if sys.version_info < (3, 9):
-    from typing_extensions import Annotated, get_args, get_origin  # pragma: no cover
-else:
-    from typing import Annotated, get_args, get_origin  # pragma: no cover
+T = TypeVar("T")
 
 __all__ = (
     "IMinuitWarning",
@@ -51,21 +55,11 @@ __all__ = (
     "make_with_signature",
     "merge_signatures",
     "describe",
+    "gradient",
+    "is_positive_definite",
 )
 
 
-class IMinuitWarning(RuntimeWarning):
-    """Generic iminuit warning."""
-
-
-class HesseFailedWarning(IMinuitWarning):
-    """HESSE failed warning."""
-
-
-class PerformanceWarning(UserWarning):
-    """Warning about performance issues."""
-
-
 class BasicView(abc.ABC):
     """
     Array-like view of parameter state.
@@ -79,7 +73,7 @@ class BasicView(abc.ABC):
     __slots__ = ("_minuit", "_ndim")
 
     def __init__(self, minuit: Any, ndim: int = 0):
-        """Not to be initialized by users."""
+        # Users should not call this __init__, instances are created by the library
         self._minuit = minuit
         self._ndim = ndim
 
@@ -116,7 +110,7 @@ class BasicView(abc.ABC):
         return self._get(index)
 
     def __setitem__(self, key: Key, value: Any) -> None:
-        """Assign a new value at key, which can be an index, parameter name, or slice."""
+        """Assign new value at key, which can be index, parameter name, or slice."""
         self._minuit._copy_state_if_needed()
         index = _key2index(self._minuit._var2pos, key)
         if isinstance(index, list):
@@ -149,7 +143,7 @@ class BasicView(abc.ABC):
         return {k: self._get(i) for i, k in enumerate(self._minuit._pos2var)}
 
 
-def _ndim(obj: Iterable) -> int:
+def _ndim(obj: Any) -> int:
     nd = 0
     while isinstance(obj, Iterable):
         nd += 1
@@ -210,7 +204,7 @@ class LimitView(BasicView):
     """Array-like view of parameter limits."""
 
     def __init__(self, minuit: Any):
-        """Not to be initialized by users."""
+        # Users should not call this __init__, instances are created by the library
         super(LimitView, self).__init__(minuit, 1)
 
     def _get(self, i: int) -> Tuple[float, float]:
@@ -245,7 +239,7 @@ class LimitView(BasicView):
         state.set_error(i, err)
 
 
-def _normalize_limit(lim: UserBound) -> Tuple[float, float]:
+def _normalize_limit(lim: Optional[Iterable]) -> Tuple[float, float]:
     if lim is None:
         return (-np.inf, np.inf)
     a, b = lim
@@ -269,8 +263,8 @@ class Matrix(np.ndarray):
 
     __slots__ = ("_var2pos",)
 
-    def __new__(cls, parameters: Union[Dict, Tuple]) -> Any:
-        """Not to be initialized by users."""
+    def __new__(cls, parameters: Union[Dict, Tuple]) -> Any:  # noqa D102
+        # Users should not call __new__, instances are created by the library
         if isinstance(parameters, dict):
             var2pos = parameters
         elif isinstance(parameters, tuple):
@@ -322,7 +316,8 @@ class Matrix(np.ndarray):
         """
         Convert matrix to dict.
 
-        Since the matrix is symmetric, the dict only contains the upper triangular matrix.
+        Since the matrix is symmetric, the dict only contains the upper triangular
+        matrix.
         """
         names = tuple(self._var2pos)
         d = {}
@@ -406,11 +401,11 @@ class FMin:
     Function minimum view.
 
     This object provides detailed metadata about the function minimum. Inspect this to
-    check what exactly happened if the fit did not converge. Use the
-    :class:`iminuit.Minuit` object to get the best fit values, their
-    uncertainties, or the function value at the minimum. For convenience, you can also
-    get a basic OK from :class:`iminuit.Minuit` with the methods
-    :attr:`iminuit.Minuit.valid` and :attr:`iminuit.Minuit.accurate`.
+    check what exactly happened if the fit did not converge. Use the attribute
+    :attr:`iminuit.Minuit.fmin` to get the best fit values, their uncertainties, or the
+    function value at the minimum. For convenience, you can also get a basic OK from
+    :class:`iminuit.Minuit` with the methods :attr:`iminuit.Minuit.valid` and
+    :attr:`iminuit.Minuit.accurate`.
 
     See Also
     --------
@@ -444,7 +439,7 @@ class FMin:
         edm_goal: float,
         time: float,
     ):
-        """Not to be initialized by users."""
+        # Users should not call this __init__, instances are created by the library
         self._src = fmin
         self._algorithm = algorithm
         self._has_parameters_at_limit = False
@@ -540,7 +535,10 @@ class FMin:
         Note: The actual verdict is computed inside the Minuit2 C++ code, so we
         cannot guarantee that is_valid is exactly equivalent to these conditions.
         """
-        return self._src.is_valid  # type:ignore
+        valid = self._src.is_valid
+        if valid and self.edm < 0:
+            return False
+        return valid
 
     @property
     def has_valid_parameters(self) -> bool:
@@ -709,7 +707,7 @@ class Param:
         self,
         *args: Union[int, str, float, Optional[Tuple[float, float]], bool],
     ):
-        """Not to be initialized by users."""
+        # Users should not call this __init__, instances are created by the library
         assert len(args) == len(self.__slots__)
         for k, arg in zip(self.__slots__, args):
             setattr(self, k, arg)
@@ -835,7 +833,8 @@ class Params(tuple):
 
 
 class MError:
-    """Minos data object.
+    """
+    Minos data object.
 
     Attributes
     ----------
@@ -890,7 +889,7 @@ class MError:
     )
 
     def __init__(self, *args: Union[int, str, float, bool]):
-        """Not to be initialized by users."""
+        # Users should not call this __init__, instances are created by the library
         assert len(args) == len(self.__slots__)
         for k, arg in zip(self.__slots__, args):
             setattr(self, k, arg)
@@ -958,60 +957,6 @@ class MErrors(OrderedDict):
         return OrderedDict.__getitem__(self, key)
 
 
-def _jacobi(
-    fn: Callable, x: NDArray, dx: NDArray, tol: float, debug: bool = False
-) -> Tuple[NDArray, NDArray]:
-    assert x.ndim == 1
-    assert dx.ndim == 1
-    assert np.all(dx >= 0)
-    assert tol > 0
-    y = fn(x)
-    yrank = np.ndim(y)
-    jac = np.zeros((1 if yrank == 0 else len(y), len(x)))
-    h = np.zeros(len(x))
-    divergence = True
-    for i, hi in enumerate(dx):
-        if i > 0:
-            h[i - 1] = 0
-        if hi == 0:
-            continue
-        h[i] = hi
-        prev_esq = np.inf
-        for iter in range(20):
-            assert h[i] > 0
-            yu = fn(x + h)
-            yd = fn(x - h)
-            du = (yu - y) / h[i]
-            dd = (y - yd) / h[i]
-            d = 0.5 * (du + dd)
-            delta = du - dd
-            if np.all(np.abs(delta) <= tol * np.abs(d)):
-                if debug:
-                    print(
-                        f"jacobi: iter={iter} converged; delta={delta} "
-                        f"threshold={tol * np.abs(d)}"
-                    )
-                jac[:, i] = d
-                break
-            esq = np.dot(delta, delta)
-            if debug:
-                print(f"jacobi: iter={iter} d={d} esq={esq} h={h}")
-            if iter > 0 and esq < prev_esq:
-                divergence = False
-            if esq >= prev_esq:
-                if divergence:
-                    print(f"jacobi: iter={iter} divergence detected")
-                    jac[:, i] = np.nan
-                else:
-                    print(f"jacobi: iter={iter} no convergence")
-                    # no convergence, use previous more accurate jac[:, i]
-                break
-            jac[:, i] = d
-            prev_esq = esq
-            h[i] *= 0.1
-    return y, jac
-
-
 @_deprecated.deprecated("use jacobi.propagate instead from jacobi library")
 def propagate(
     fn: Callable,
@@ -1021,8 +966,8 @@ def propagate(
     """
     Numerically propagates the covariance into a new space.
 
-    This function is deprecated and will be removed. Please use jacobi.propagate from the
-    jacobi library, which is more accurate. The signatures of the two functions are
+    This function is deprecated and will be removed. Please use jacobi.propagate from
+    the jacobi library, which is more accurate. The signatures of the two functions are
     compatible, so it is a drop-in replacement.
 
     Parameters
@@ -1040,17 +985,20 @@ def propagate(
         y is the result of fn(x)
         ycov is the propagated covariance matrix.
     """
+    from scipy.optimize import approx_fprime
+
     vx = np.atleast_1d(x)  # type:ignore
     if np.ndim(cov) != 2:  # type:ignore
         raise ValueError("cov must be 2D array-like")
     vcov = np.atleast_2d(cov)  # type:ignore
     if vcov.shape[0] != vcov.shape[1]:
         raise ValueError("cov must have shape (N, N)")
-    tol = 1e-2
+    tol = 1e-10
     dx = (np.diag(vcov) * tol) ** 0.5
     if not np.all(dx >= 0):
         raise ValueError("diagonal elements of covariance matrix must be non-negative")
-    y, jac = _jacobi(fn, vx, dx, tol)
+    y = fn(vx)
+    jac = np.atleast_2d(approx_fprime(vx, fn, dx))
     ycov = np.einsum("ij,kl,jl", jac, jac, vcov)
     return y, np.squeeze(ycov) if np.ndim(y) == 0 else ycov
 
@@ -1067,8 +1015,8 @@ class _Timer:
 
 
 @_deprecated.deprecated(
-    "Use of `func_code` attribute to declare parameters is deprecated. "
-    "Use `_parameters` instead, which is a dict of parameter names to limits."
+    "Use of ``func_code`` attribute to declare parameters is deprecated. "
+    "Use ``_parameters`` instead, which is a dict of parameter names to limits."
 )
 def make_func_code(params: Collection[str]) -> Namespace:
     """
@@ -1091,6 +1039,8 @@ def make_with_signature(
 
     Parameters
     ----------
+    callable:
+        Original callable with positional arguments, whose names shall be changed.
     *varnames: sequence of str
         Replace the first N argument names with these.
     **replacements: mapping of str to str, optional
@@ -1098,7 +1048,7 @@ def make_with_signature(
 
     Returns
     -------
-    callable with new argument names.
+    Callable with positional-only arguments.
     """
     pars = describe(callable, annotations=True)
     if pars:
@@ -1116,7 +1066,7 @@ def make_with_signature(
         def __init__(self, parameters):
             self._parameters = parameters
 
-        def __call__(self, *args: object) -> object:
+        def __call__(self, *args: Any) -> Any:
             return callable(*args)
 
     return Caller(pars)
@@ -1124,7 +1074,7 @@ def make_with_signature(
 
 def merge_signatures(
     callables: Iterable[Callable], annotations: bool = False
-) -> Tuple[Any, List[Tuple[int, ...]]]:
+) -> Tuple[Any, List[List[int]]]:
     """
     Merge signatures of callables with positional arguments.
 
@@ -1136,12 +1086,15 @@ def merge_signatures(
 
         parameters, mapping = merge_signatures(f, g)
         # parameters is ('x', 'y', 'z', 'p')
-        # mapping is ((0, 1, 2), (0, 3))
+        # mapping is ([0, 1, 2], [0, 3])
 
     Parameters
     ----------
-    callable : callable
-        Callable whose parameters can be extracted with :func:`describe`.
+    callables :
+        Callables whose parameters can be extracted with :func:`describe`.
+    annotations : bool, optional
+        Whether to return the annotions. Default is False, for backward
+        compatibility.
 
     Returns
     -------
@@ -1163,7 +1116,7 @@ def merge_signatures(
                 amap.append(len(args))
                 args.append(k)
                 anns.append(ann)
-        mapping.append(tuple(amap))
+        mapping.append(amap)
 
     if annotations:
         return {k: a for (k, a) in zip(args, anns)}, mapping
@@ -1171,15 +1124,13 @@ def merge_signatures(
 
 
 @overload
-def describe(callable: Callable) -> List[str]:
-    ...  # pragma: no cover
+def describe(callable: Callable) -> List[str]: ...  # pragma: no cover
 
 
 @overload
 def describe(
     callable: Callable, annotations: bool
-) -> Dict[str, Optional[Tuple[float, float]]]:
-    ...  # pragma: no cover
+) -> Dict[str, Optional[Tuple[float, float]]]: ...  # pragma: no cover
 
 
 def describe(callable, *, annotations=False):
@@ -1189,7 +1140,7 @@ def describe(callable, *, annotations=False):
     Parameters
     ----------
     callable : callable
-        Callable whose parameters should be extracted.
+        Callable whose parameter names should be extracted.
     annotations : bool, optional
         Whether to also extract annotations. Default is false.
 
@@ -1206,10 +1157,10 @@ def describe(callable, *, annotations=False):
     Parameter names are extracted with the following three methods, which are attempted
     in order. The first to succeed determines the result.
 
-    1.  Using ``obj._parameters``, which is a dict that makes parameter names to parameter
-        limits or None if the parameter has to limits. Users are encouraged to use this
-        mechanism to provide signatures for objects that otherwise would not have a
-        detectable signature. Example::
+    1.  Using ``obj._parameters``, which is a dict that maps parameter names to
+        parameter limits or None if the parameter has to limits. Users are encouraged
+        to use this mechanism to provide signatures for objects that otherwise would
+        not have a detectable signature. Example::
 
             def f(*args):  # no signature
                 x, y = args
@@ -1217,16 +1168,16 @@ def describe(callable, *, annotations=False):
 
             f._parameters = {"x": None, "y": (1, 4)}
 
-        Here, the first parameter is declared to have no limits (values from minus to plus
-        infinity are allowed), while the second parameter is declared to have limits, it
-        cannot be smaller than 1 or larger than 4.
+        Here, the first parameter is declared to have no limits (values from minus to
+        plus infinity are allowed), while the second parameter is declared to have
+        limits, it cannot be smaller than 1 or larger than 4.
 
         Note: In the past, the ``func_code`` attribute was used for a similar purpose as
         ``_parameters``. It is still supported for legacy code, but should not be used
         anymore in new code, since it does not support declaring parameter limits.
 
-        If an objects has a ``func_code`` attribute, it is used to detect the parameters.
-        Example::
+        If an objects has a ``func_code`` attribute, it is used to detect the
+        parameters. Example::
 
             from iminuit.util import make_func_code
 
@@ -1245,10 +1196,10 @@ def describe(callable, *, annotations=False):
                 def __call__(self, a, b):
                     ...
 
-        Limits are supported via annotations, using the Annotated type that was introduced
-        in Python-3.9 and can be obtained from the external package typing_extensions in
-        Python-3.8. In the following example, the second parameter has a lower limit at
-        0, because it must be positive::
+        Limits are supported via annotations, using the Annotated type that was
+        introduced in Python-3.9 and can be obtained from the external package
+        typing_extensions in Python-3.8. In the following example, the second parameter
+        has a lower limit at 0, because it must be positive::
 
             from typing import Annotated
 
@@ -1256,33 +1207,33 @@ def describe(callable, *, annotations=False):
                 ...
 
         There are no standard annotations yet at the time of this writing. iminuit
-        supports the annotations Gt, Ge, Lt, Le from the external package annotated-types,
-        and interprets slice notation as limits.
+        supports the annotations Gt, Ge, Lt, Le from the external package
+        annotated-types, and interprets slice notation as limits.
 
     3.  Using the docstring. The docstring is textually parsed to detect the parameter
         names. This requires that a docstring is present which follows the Python
-        standard formatting for function signatures.
+        standard formatting for function signatures. Here is a contrived example which
+        is parsed correctly: fn(a, b: float, int c, d=1, e: float=2.2, double e=3).
 
     Ambiguous cases with positional and keyword argument are handled in the following
     way::
 
+        def fcn(a, b, *args, **kwargs): ...
         # describe returns [a, b];
         # *args and **kwargs are ignored
-        def fcn(a, b, *args, **kwargs): ...
 
+        def fcn(a, b, c=1): ...
         # describe returns [a, b, c];
         # positional arguments with default values are detected
-        def fcn(a, b, c=1): ...
-
     """
     if _address_of_cfunc(callable) != 0:
         return {} if annotations else []
 
     args = (
         getattr(callable, "_parameters", {})
-        or _arguments_from_func_code(callable)
-        or _parameters_from_inspect(callable)
-        or _arguments_from_docstring(callable)
+        or _describe_impl_func_code(callable)
+        or _describe_impl_inspect(callable)
+        or _describe_impl_docstring(callable)
     )
 
     if annotations:
@@ -1291,65 +1242,16 @@ def describe(callable, *, annotations=False):
     return list(args)
 
 
-def _arguments_from_func_code(callable):
+def _describe_impl_func_code(callable):
     # Check (faked) f.func_code; for backward-compatibility with iminuit-1.x
     if hasattr(callable, "func_code"):
-        warnings.warn(
-            "Using the `func_code` attribute to dynamically declare parameter names is "
-            "deprecated, use the attribute `_parameters` instead (a dict of strings "
-            "to limits)",
-            np.VisibleDeprecationWarning,
-            stacklevel=1,
-        )
+        # cannot warn about deprecation here, since numba.njit also uses .func_code
         fc = callable.func_code
         return {x: None for x in fc.co_varnames[: fc.co_argcount]}
     return {}
 
 
-def _get_limit(annotation: Union[type, Annotated[float, Any], str]):
-    from iminuit import typing
-
-    if isinstance(annotation, str):
-        # This provides limited support for string annotations, which
-        # have a lot of problems, see https://peps.python.org/pep-0649.
-        try:
-            annotation = eval(annotation, None, typing.__dict__)
-        except NameError:
-            # We ignore unknown annotations to fix issue #846.
-            # I cannot replicate here what inspect.signature(..., eval_str=True) does.
-            # I need a dict with the global objects at the call site of describe, but
-            # it is not globals(). Anyway, when using strings, only the annotations
-            # from the package annotated-types are supported.
-            pass
-
-    if annotation == inspect.Parameter.empty:
-        return None
-
-    if get_origin(annotation) is not Annotated:
-        return None
-
-    tp, *constraints = get_args(annotation)
-    assert tp is float
-    lim = [-np.inf, np.inf]
-    for c in constraints:
-        if isinstance(c, slice):
-            if c.start is not None:
-                lim[0] = c.start
-            if c.stop is not None:
-                lim[1] = c.stop
-        # Minuit does not distinguish between closed and open intervals
-        elif hasattr(c, "gt"):
-            lim[0] = c.gt
-        elif hasattr(c, "ge"):
-            lim[0] = c.ge
-        elif hasattr(c, "lt"):
-            lim[1] = c.lt
-        elif hasattr(c, "le"):
-            lim[1] = c.le
-    return tuple(lim)
-
-
-def _parameters_from_inspect(callable):
+def _describe_impl_inspect(callable):
     try:
         signature = inspect.signature(callable)
     except ValueError:  # raised when used on built-in function
@@ -1367,7 +1269,7 @@ def _parameters_from_inspect(callable):
     return r
 
 
-def _arguments_from_docstring(callable):
+def _describe_impl_docstring(callable):
     doc = inspect.getdoc(callable)
 
     if doc is None:
@@ -1377,6 +1279,7 @@ def _arguments_from_docstring(callable):
     #   min(iterable, *[, default=obj, key=func]) -> value
     #   min(arg1, arg2, *args, *[, key=func]) -> value
     #   Foo.bar(self, int ncall_me =10000, [resume=True, int nsplit=1])
+    #   Foo.baz(self: Foo, ncall_me: int =10000)
 
     try:
         # function wrapper functools.partial does not offer __name__,
@@ -1402,12 +1305,13 @@ def _arguments_from_docstring(callable):
     items = [x.strip(" []") for x in doc[start : start + ich].split(",")]
 
     # strip self if callable is a class method
-    if items[0] == "self":
+    if inspect.ismethod(callable):
         items = items[1:]
 
     #   "iterable", "*", "default=obj", "key=func"
     #   "arg1", "arg2", "*args", "*", "key=func"
     #   "int ncall_me =10000", "resume=True", "int nsplit=1"
+    #   "ncall_me: int =10000"
 
     try:
         i = items.index("*args")
@@ -1418,22 +1322,86 @@ def _arguments_from_docstring(callable):
     #   "iterable", "*", "default=obj", "key=func"
     #   "arg1", "arg2", "*", "key=func"
     #   "int ncall_me =10000", "resume=True", "int nsplit=1"
+    #   "ncall_me: int =10000"
 
     def extract(s: str) -> str:
-        a = s.find(" ")
-        b = s.find("=")
-        if a < 0:
-            a = 0
-        if b < 0:
-            b = len(s)
-        return s[a:b].strip()
+        i = s.find("=")
+        if i >= 0:
+            s = s[:i]
+        i = s.find(":")
+        if i >= 0:
+            return s[:i].strip()
+        s = s.strip()
+        i = s.find(" ")
+        if i >= 0:
+            return s[i:].strip()
+        return s
 
-    #   "iterable", "default", "key"
+    #   "iterable", "*", "default", "key"
     #   "arg1", "arg2", "key"
     #   "ncall_me", "resume", "nsplit"
+    #   "ncall_me"
     return {extract(x): None for x in items if x != "*"}
 
 
+def _get_limit(
+    annotation: Union[type, Annotated[float, Any], str],
+) -> Optional[Tuple[float, float]]:
+    from iminuit import typing
+
+    if isinstance(annotation, str):
+        # This provides limited support for string annotations, which
+        # have a lot of problems, see https://peps.python.org/pep-0649.
+        try:
+            annotation = eval(annotation, None, typing.__dict__)
+        except NameError:
+            # We ignore unknown annotations to fix issue #846.
+            # I cannot replicate here what inspect.signature(..., eval_str=True) does.
+            # I need a dict with the global objects at the call site of describe, but
+            # it is not globals(). Anyway, when using strings, only the annotations
+            # from the package annotated-types are supported.
+            pass
+
+    if annotation == inspect.Parameter.empty:
+        return None
+
+    if get_origin(annotation) is not Annotated:
+        return None
+
+    tp, *constraints = get_args(annotation)
+    assert tp is float
+    lower = -np.inf
+    upper = np.inf
+    for c in constraints:
+        if isinstance(c, slice):
+            if c.start is not None:
+                lower = c.start
+            if c.stop is not None:
+                upper = c.stop
+            continue
+        if isinstance(c, Sequence):
+            lower, upper = c
+            continue
+
+        # Minuit does not distinguish between closed and open intervals.
+        # We use a chain of ifs so that the code also works with the
+        # `Interval` class, which contains several of those attributes
+        # and which can be None.
+        gt = getattr(c, "gt", None)
+        ge = getattr(c, "ge", None)
+        lt = getattr(c, "lt", None)
+        le = getattr(c, "le", None)
+        if gt is not None:
+            lower = gt
+        if ge is not None:
+            lower = ge
+        if lt is not None:
+            upper = lt
+        if le is not None:
+            upper = le
+    return lower, upper
+
+
 def _guess_initial_step(val: float) -> float:
     return 1e-2 * abs(val) if val != 0 else 1e-1  # heuristic
 
@@ -1460,6 +1428,8 @@ def _key2index(
     var2pos: Dict[str, int],
     key: Key,
 ) -> Union[int, List[int]]:
+    if key is ...:
+        return list(range(len(var2pos)))
     if isinstance(key, slice):
         return _key2index_from_slice(var2pos, key)
     if not isinstance(key, str) and isinstance(key, Iterable):
@@ -1496,16 +1466,25 @@ def _iterate(x):
             yield xi
 
 
-def _replace_none(x, v):
+def _replace_none(x: Optional[T], v: T) -> T:
     if x is None:
         return v
     return x
 
 
-# poor-mans progressbar if progressbar2 is not available
-class _ProgressBar:
-    def _update(self, v):
-        self._out.write(f"\r{100 * v:.0f} %")
+# poor-mans progressbar
+class ProgressBar:
+    """
+    Simple progress bar.
+
+    Renders as nice HTML progressbar in Jupyter notebooks.
+    """
+
+    value: int = 0
+    max_value: int
+
+    def _update(self, fraction):
+        self._out.write(f"\r{100 * fraction:.0f} %")
         self._out.flush()
 
     def _finish(self):
@@ -1513,6 +1492,14 @@ class _ProgressBar:
         self._out.flush()
 
     def __init__(self, max_value):
+        """
+        Initialize bar.
+
+        Parameters
+        ----------
+        max_value: int
+            Total number of entries.
+        """
         self.max_value = max_value
         self._out = sys.stdout
 
@@ -1536,15 +1523,18 @@ class _ProgressBar:
         except ModuleNotFoundError:
             pass
 
+        self._update(self.value)
+
     def __enter__(self, *args):
-        self.value = 0
-        self._update(0)
+        """Noop."""
         return self
 
     def __exit__(self, *args):
+        """Clean up the bar."""
         self._finish()
 
     def __add__(self, v):
+        """Increment progress."""
         self.value += v
         self._update(self.value / self.max_value)
         return self
@@ -1574,7 +1564,8 @@ def _histogram_segments(mask, xe, masked):
     return segments
 
 
-def _smart_sampling(f, xmin, xmax, start=5, tol=5e-3):
+def _smart_sampling(f, xmin, xmax, start=20, tol=5e-3, maxiter=20, maxtime=10):
+    t0 = monotonic()
     x = np.linspace(xmin, xmax, start)
     ynew = f(x)
     ymin = np.min(ynew)
@@ -1582,10 +1573,23 @@ def _smart_sampling(f, xmin, xmax, start=5, tol=5e-3):
     y = {xi: yi for (xi, yi) in zip(x, ynew)}
     a = x[:-1]
     b = x[1:]
+    niter = 0
     while len(a):
-        if len(y) > 10000:
-            warnings.warn("Too many points", RuntimeWarning)  # pragma: no cover
-            break  # pragma: no cover
+        niter += 1
+        if niter > maxiter:
+            msg = (
+                f"Iteration limit {maxiter} in smart sampling reached, "
+                f"produced {len(y)} points"
+            )
+            warnings.warn(msg, RuntimeWarning)
+            break
+        if monotonic() - t0 > maxtime:
+            msg = (
+                f"Time limit {maxtime} in smart sampling reached, "
+                f"produced {len(y)} points"
+            )
+            warnings.warn(msg, RuntimeWarning)
+            break
         xnew = 0.5 * (a + b)
         ynew = f(xnew)
         ymin = min(ymin, np.min(ynew))
@@ -1597,10 +1601,11 @@ def _smart_sampling(f, xmin, xmax, start=5, tol=5e-3):
             + np.fromiter((y[bi] for bi in b), float)
         )
         dy = np.abs(ynew - yint)
+        dx = np.abs(b - a)
 
-        mask = dy > tol * (ymax - ymin)
-
-        # intervals which do not pass interpolation test
+        # in next iteration, handle intervals which do not
+        # pass interpolation test and are not too narrow
+        mask = (dy > tol * (ymax - ymin)) & (dx > tol * abs(xmax - xmin))
         a = a[mask]
         b = b[mask]
         xnew = xnew[mask]
@@ -1610,3 +1615,72 @@ def _smart_sampling(f, xmin, xmax, start=5, tol=5e-3):
     xy = list(y.items())
     xy.sort()
     return np.transpose(xy)
+
+
+def _detect_log_spacing(x: NDArray) -> bool:
+    # x should never contain NaN
+    x = np.sort(x)
+    if x[0] <= 0:
+        return False
+    d_lin = np.diff(x)
+    d_log = np.diff(np.log(x))
+    lin_rel_std = np.std(d_lin) / np.mean(d_lin)
+    log_rel_std = np.std(d_log) / np.mean(d_log)
+    return log_rel_std < lin_rel_std
+
+
+def gradient(fcn: Cost) -> Optional[CostGradient]:
+    """
+    Return a callable which computes the gradient of fcn or None.
+
+    Parameters
+    ----------
+    fcn: Cost
+        Cost function which may provide a callable gradient. How the gradient
+        is detected is specified below in the Notes.
+
+    Notes
+    -----
+    This function checks whether the following attributes exist: `fcn.grad` and
+    `fcn.has_grad`. If `fcn.grad` exists and is a CostGradient, it is returned unless
+    `fcn.has_grad` exists and is False. If no useable gradient is detected, None is
+    returned.
+
+    Returns
+    -------
+    callable or None
+        The gradient function or None
+    """
+    grad = getattr(fcn, "grad", None)
+    has_grad = getattr(fcn, "has_grad", True)
+    if grad and isinstance(grad, CostGradient) and has_grad:
+        return grad
+    return None
+
+
+def is_positive_definite(m: ArrayLike) -> bool:
+    """
+    Return True if argument is a positive definite matrix.
+
+    This test is somewhat expensive, because we attempt a cholesky decomposition.
+
+    Parameters
+    ----------
+    m : array-like
+        Matrix to be tested.
+
+    Returns
+    -------
+    bool
+        True if matrix is positive definite and False otherwise.
+    """
+    m = np.atleast_2d(m)
+    if np.all(m.T == m):
+        # maybe check this first https://en.wikipedia.org/wiki/Diagonally_dominant_matrix
+        # and only try cholesky if that fails
+        try:
+            np.linalg.cholesky(m)
+        except np.linalg.LinAlgError:
+            return False
+        return True
+    return False
